/*
 * AVC8000 Copyright (C) 2012 Advanced Micro Peripherals Ltd
 * Driver for AVC8000 cards
 *
 * (c) 2009-10 liran <jlee@techwellinc.com.cn> [Techwell China]
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Thanks to yiliang for variable audio packet length and more audio
 *  formats support.
 */

#include <sound/core.h>
#include <sound/control.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/initval.h>

#include "avc8000.h"
#include "avc8000-device.h"
#include "avc8000-reg.h"

#if(LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0))
#include <linux/vmalloc.h>
#endif

// just use the one in core
//MODULE_DESCRIPTION("alsa driver module for AVC8000 cards");
//MODULE_AUTHOR("AMP");
//MODULE_LICENSE("GPL");

static long avc8k_audio_nr = 0;

/*
 * PCM structure
 */

typedef struct snd_card_avc8k_pcm {
	struct avc8k_adev *dev;

	spinlock_t lock;

	struct snd_pcm_substream *substream;
} snd_card_avc8k_pcm_t;


/*
 * avc8k audio DMA IRQ handler
 *
 *   Called whenever we get an avc8k audio interrupt
 *   Handles shifting between the 2 buffers, manages the read counters,
 *   and notifies ALSA when periods elapse
 *
 */

static void avc8k_audio_advance_one_period(struct avc8k_adev *dev)
{
  dev->read_offset = dev->period_idx * dev->blksize;
  dev->period_idx ++;
  if (dev->period_idx == dev->blocks)
    dev->period_idx = 0;

  // configure the next DMA buffer
  // period_idx - 1, just transferred,
  // period_idx,     being transferred
  // period_idx + 1, next buffer
  dev->dma_blk_addr += dev->blksize;
  if (dev->period_idx + 1 == dev->blocks)
    dev->dma_blk_addr -= dev->bufsize;
}

void avc8k_audio_irq(struct avc8k_adev *dev, u32 dma_status, u32 pb_status)
{
  int pb = (pb_status>>(dev->channel_id+AVC8K68XX_AUDIO_BEGIN)) & 1;

  daprintk(DPRT_LEVEL0, dev, "%s(%d)\n", __func__, pb);

  /* next block addr */

  /* update status & wake waiting readers */
  spin_lock(&dev->slock);
  avc8k_audio_advance_one_period(dev);
  spin_unlock(&dev->slock);

  if (dev->pb_flag != pb)
  {
    daprintk(1, dev, "pb flag mismatch\n");
    spin_lock(&dev->slock);
    avc8k_dev_set_adma_buffer (dev, dev->dma_blk_addr, dev->pb_flag);
    avc8k_audio_advance_one_period(dev);
    dev->pb_flag = 1 - dev->pb_flag;
    spin_unlock(&dev->slock);
    if(NULL == (dev->substream))
	{
		daprintk(1, dev, "avc8k_audio_irq: null pointer!\n");
	}
	else
	{
		snd_pcm_period_elapsed(dev->substream);
	}
  }
  else
  {
	  spin_lock(&dev->slock);
	  avc8k_dev_set_adma_buffer (dev, dev->dma_blk_addr, dev->pb_flag);
	  dev->pb_flag = 1 - dev->pb_flag;
	  spin_unlock(&dev->slock);
  }
  snd_pcm_period_elapsed(dev->substream);
}

/*
 * ALSA capture trigger
 *
 *   - One of the ALSA capture callbacks.
 *
 *   Called whenever a capture is started or stopped. Must be defined,
 *   but there's nothing we want to do here
 *
 */

static int snd_card_avc8k_capture_trigger(struct snd_pcm_substream * substream,
					  int cmd)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	snd_card_avc8k_pcm_t *pcm = runtime->private_data;
	struct avc8k_adev *dev=pcm->dev;
	int err = 0;

    daprintk(DPRT_LEVEL0, dev, "%s(%d)\n", __func__, (cmd == SNDRV_PCM_TRIGGER_START));

	spin_lock(&dev->slock);
	if (cmd == SNDRV_PCM_TRIGGER_START) {
		/* start dma */
		avc8k_dev_run_audio(dev, true);
	} else if (cmd == SNDRV_PCM_TRIGGER_STOP) {
		/* stop dma */
		avc8k_dev_run_audio(dev, false);
	} else {
		err = -EINVAL;
	}
	spin_unlock(&dev->slock);

	return err;
}

#if(LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0))
static int avc8k_alsa_dma_init(struct avc8k_adev *dev, int nr_pages)
{
		struct avc8k_dmasound *dma = &dev->dmasound;
        struct page *pg;
        int i;

        dma->vaddr = vmalloc_32(nr_pages << PAGE_SHIFT);
        if (NULL == dma->vaddr) {
                printk("vmalloc_32(%d pages) failed\n", nr_pages);
                return -ENOMEM;
        }

        printk("vmalloc is at addr 0x%08lx, size=%d\n",
                                (unsigned long)dma->vaddr,
                                nr_pages << PAGE_SHIFT);

        memset(dma->vaddr, 0, nr_pages << PAGE_SHIFT);
        dma->nr_pages = nr_pages;

        dma->sglist = vzalloc(dma->nr_pages * sizeof(*dma->sglist));
        if (NULL == dma->sglist)
                goto vzalloc_err;

        sg_init_table(dma->sglist, dma->nr_pages);
        for (i = 0; i < dma->nr_pages; i++) {
                pg = vmalloc_to_page(dma->vaddr + i * PAGE_SIZE);
                if (NULL == pg)
                        goto vmalloc_to_page_err;
                sg_set_page(&dma->sglist[i], pg, PAGE_SIZE, 0);
        }
        return 0;

vmalloc_to_page_err:
        vfree(dma->sglist);
        dma->sglist = NULL;
vzalloc_err:
        vfree(dma->vaddr);
        dma->vaddr = NULL;
        return -ENOMEM;
}

static int avc8k_alsa_dma_map(struct avc8k_adev *dev)
{
		struct avc8k_dmasound *dma = &dev->dmasound;

        dma->sglen = dma_map_sg(&dev->chip->pci->dev, dma->sglist,
                         dma->nr_pages, PCI_DMA_FROMDEVICE);

        if (0 == dma->sglen) {
                pr_warn("%s: saa7134_alsa_map_sg failed\n", __func__);
                return -ENOMEM;
        }
        return 0;
}

static int avc8k_alsa_dma_unmap(struct avc8k_adev *dev)
{
		struct avc8k_dmasound *dma = &dev->dmasound;

        if (!dma->sglen)
                return 0;

        dma_unmap_sg(&dev->chip->pci->dev, dma->sglist, dma->sglen, PCI_DMA_FROMDEVICE);
        dma->sglen = 0;
        return 0;
}

static int avc8k_alsa_dma_free(struct avc8k_dmasound *dma)
{
        vfree(dma->sglist);
        dma->sglist = NULL;
        vfree(dma->vaddr);
        dma->vaddr = NULL;
        return 0;
}
#endif

/*
 * DMA buffer initialization
 *
 *   Uses V4L functions to initialize the DMA. Shouldn't be necessary in
 *  ALSA, but I was unable to use ALSA's own DMA, and had to force the
 *  usage of V4L's
 *
 *   - Copied verbatim from saa7134-oss.
 *
 */

static int dsp_buffer_init(struct avc8k_adev *dev)
{
	int err;

	BUG_ON(!dev->bufsize);

#if(LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0))
	err = avc8k_alsa_dma_init(dev,
                           (dev->bufsize + PAGE_SIZE) >> PAGE_SHIFT);
    if (0 != err)
        return err;
#else
	videobuf_dma_init(&dev->dma);
	err = videobuf_dma_init_kernel(&dev->dma, PCI_DMA_FROMDEVICE,
				       (dev->bufsize + PAGE_SIZE) >> PAGE_SHIFT);
#endif

	return err;
}

/*
 * DMA buffer release
 *
 *   Called after closing the device, during snd_card_avc8k_capture_close
 *
 */

static int dsp_buffer_free(struct avc8k_adev *dev)
{
	BUG_ON(!dev->blksize);

#if(LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0))
	avc8k_alsa_dma_free(&dev->dmasound);
#else
	videobuf_dma_free(&dev->dma);
#endif
	dev->blocks  = 0;
	dev->blksize = 0;
	dev->bufsize = 0;

	return 0;
}

/*
 * ALSA PCM preparation
 *
 *   - One of the ALSA capture callbacks.
 *
 *   Called right after the capture device is opened, this function configures
 *  the buffer using the previously defined functions, allocates the memory,
 *  sets up the hardware registers, and then starts the DMA. When this function
 *  returns, the audio should be flowing.
 *
 */

static int snd_card_avc8k_capture_prepare(struct snd_pcm_substream * substream)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	snd_card_avc8k_t *card_avc8k = snd_pcm_substream_chip(substream);
	struct avc8k_adev *dev = card_avc8k->dev;
	snd_card_avc8k_pcm_t *pcm = runtime->private_data;

    daprintk(DPRT_LEVEL0, dev, "%s(%d, %d, %d)\n", __func__, runtime->rate, runtime->channels, runtime->sample_bits);

	pcm->dev->substream = substream;

    daprintk(DPRT_LEVEL0, dev, "%s_%d, rate %d, channels %d, depth %d\n",
         __func__,
         dev->channel_id,
         runtime->rate,
         runtime->channels,
         runtime->sample_bits);

    avc8k_dev_set_audio(dev->chip, runtime->rate, runtime->sample_bits, runtime->channels, dev->blksize);

	dev->rate = runtime->rate;

	return 0;
}

/*
 * ALSA pointer fetching
 *
 *   - One of the ALSA capture callbacks.
 *
 *   Called whenever a period elapses, it must return the current hardware
 *  position of the buffer.
 *   Also resets the read counter used to prevent overruns
 *
 */

static snd_pcm_uframes_t
snd_card_avc8k_capture_pointer(struct snd_pcm_substream * substream)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	snd_card_avc8k_pcm_t *pcm = runtime->private_data;
	struct avc8k_adev *dev=pcm->dev;

    daprintk(DPRT_LEVEL0, dev, "%s(%d)\n", __func__, dev->read_offset);

	return bytes_to_frames(runtime, dev->read_offset);
}

/*
 * ALSA hardware capabilities definition
 */

static struct snd_pcm_hardware snd_card_avc8k_capture =
{
	.info =     (SNDRV_PCM_INFO_MMAP | SNDRV_PCM_INFO_INTERLEAVED |
				 SNDRV_PCM_INFO_BLOCK_TRANSFER |
				 SNDRV_PCM_INFO_MMAP_VALID),
  .formats = SNDRV_PCM_FMTBIT_S16_LE | SNDRV_PCM_FMTBIT_S8,
  .rates = SNDRV_PCM_RATE_KNOT,
  .rate_min = 8000,
  .rate_max = 48000,
  .channels_min = 1,
  .channels_max = 1,
  .buffer_bytes_max =	4096,
  .period_bytes_min =	256,
  .period_bytes_max =	1024,
  .periods_min      = 4,
  .periods_max      = 8,
};

static void snd_card_avc8k_runtime_free(struct snd_pcm_runtime *runtime)
{
	snd_card_avc8k_pcm_t *pcm = runtime->private_data;
	struct avc8k_adev *dev = pcm->dev;

    daprintk(DPRT_LEVEL0, dev, "%s()\n", __func__);

	kfree(pcm);
}


/*
 * ALSA hardware params
 *
 *   - One of the ALSA capture callbacks.
 *
 *   Called on initialization, right before the PCM preparation
 *
 */

static int snd_card_avc8k_hw_params(struct snd_pcm_substream * substream,
				      struct snd_pcm_hw_params * hw_params)
{
	snd_card_avc8k_t *card_avc8k = snd_pcm_substream_chip(substream);
	struct avc8k_adev *dev = card_avc8k->dev;
	unsigned int period_size, periods;
	struct scatterlist *list;
	int err;

	period_size = params_period_bytes(hw_params);
	periods = params_periods(hw_params);

	if (period_size < 0x100 || period_size > 0x10000)
		return -EINVAL;
	if (periods < 2)
		return -EINVAL;
	if (period_size * periods > 1024 * 1024)
		return -EINVAL;

    daprintk(DPRT_LEVEL0, dev, "%s(bufsize=%d)\n", __func__, period_size * periods);

	if (dev->blocks == periods &&
	    dev->blksize == period_size)
		return 0;

	/* release the old buffer */
	if (substream->runtime->dma_area) {
#if(LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0))
		avc8k_pgtable_free(dev->chip->pci, &dev->dmasound.pt);
		avc8k_alsa_dma_unmap(dev);
#elif((LINUX_VERSION_CODE > KERNEL_VERSION(2,6,35))||(RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(6,3)))
		videobuf_dma_unmap(&dev->chip->pci->dev, &dev->dma);
#else
		videobuf_sg_dma_unmap(&dev->chip->pci->dev, &dev->dma);
#endif
		dsp_buffer_free(dev);
		substream->runtime->dma_area = NULL;
	}

    dev->pb_flag    = 0;
    dev->period_idx = 0;
	dev->blocks  = periods;
	dev->blksize = period_size;
	dev->bufsize = period_size * periods;

	err = dsp_buffer_init(dev);
	if (0 != err) {
		dev->blocks  = 0;
		dev->blksize = 0;
		dev->bufsize = 0;
		return err;
	}

#if(LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0))
	if (0 != (err = avc8k_alsa_dma_map(dev))) {
#elif(LINUX_VERSION_CODE > KERNEL_VERSION(2,6,35))||(RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(6,3))
	if (0 != (err = videobuf_dma_map(&dev->chip->pci->dev, &dev->dma))) {
#else
	if (0 != (err = videobuf_sg_dma_map(&dev->chip->pci->dev, &dev->dma))) {

#endif
		dsp_buffer_free(dev);
		return err;
	}

#if(LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0))
	list = dev->dmasound.sglist;

    daprintk(DPRT_LEVEL0, dev, "%s_%d: period_size %d, periods %d, sglen %d.\n",
         __func__,
         dev->channel_id,
         period_size,
         periods,
         dev->dmasound.sglen);
#else
    list = dev->dma.sglist;

    daprintk(DPRT_LEVEL0, dev, "%s_%d: period_size %d, periods %d, sglen %d.\n",
         __func__,
         dev->channel_id,
         period_size,
         periods,
         dev->dma.sglen);
#endif

    dev->dma_blk_addr = cpu_to_le32(sg_dma_address(list) - list->offset);
    avc8k_dev_set_adma_buffer (dev, dev->dma_blk_addr, 0);

    dev->dma_blk_addr += dev->blksize;
    avc8k_dev_set_adma_buffer (dev, dev->dma_blk_addr, 1);

	/* I should be able to use runtime->dma_addr in the control
	   byte, but it doesn't work. So I allocate the DMA using the
	   V4L functions, and force ALSA to use that as the DMA area */

#if(LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0))
	substream->runtime->dma_area = dev->dmasound.vaddr;
#elif((LINUX_VERSION_CODE > KERNEL_VERSION(2,6,35))||(RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(6,3)))
	substream->runtime->dma_area = dev->dma.vaddr;
#else
	substream->runtime->dma_area = dev->dma.vmalloc;

#endif
	substream->runtime->dma_bytes = dev->bufsize;
	substream->runtime->dma_addr = 0;

	return 0;
}

/*
 * ALSA hardware release
 *
 *   - One of the ALSA capture callbacks.
 *
 *   Called after closing the device, but before snd_card_avc8k_capture_close
 *   It stops the DMA audio and releases the buffers.
 *
 */

static int snd_card_avc8k_hw_free(struct snd_pcm_substream * substream)
{
	snd_card_avc8k_t *card_avc8k = snd_pcm_substream_chip(substream);
	struct avc8k_adev *dev;

	dev = card_avc8k->dev;

    daprintk(DPRT_LEVEL0, dev, "%s()\n", __func__);

	if (substream->runtime->dma_area) {
#if(LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0))
   		avc8k_pgtable_free(dev->chip->pci, &dev->dmasound.pt);
   		avc8k_alsa_dma_unmap(dev);
#elif((LINUX_VERSION_CODE > KERNEL_VERSION(2,6,35))||(RHEL_RELEASE_CODE >= RHEL_RELEASE_VERSION(6,3)))
		videobuf_dma_unmap(&dev->chip->pci->dev, &dev->dma);
#else
		videobuf_sg_dma_unmap(&dev->chip->pci->dev, &dev->dma);

#endif
		dsp_buffer_free(dev);
		substream->runtime->dma_area = NULL;
	}

	return 0;
}

/*
 * ALSA capture finish
 *
 *   - One of the ALSA capture callbacks.
 *
 *   Called after closing the device.
 *
 */

static int snd_card_avc8k_capture_close(struct snd_pcm_substream * substream)
{
	snd_card_avc8k_t *card_avc8k = snd_pcm_substream_chip(substream);
	struct avc8k_adev *dev = card_avc8k->dev;

    daprintk(DPRT_LEVEL0, dev, "%s()\n", __func__);

	if (card_avc8k->mute_was_on) {
	}
	return 0;
}

/*
 * ALSA capture start
 *
 *   - One of the ALSA capture callbacks.
 *
 *   Called when opening the device. It creates and populates the PCM
 *  structure
 *
 */

static int snd_card_avc8k_capture_open(struct snd_pcm_substream * substream)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	snd_card_avc8k_pcm_t *pcm;
	snd_card_avc8k_t *card_avc8k = snd_pcm_substream_chip(substream);
	struct avc8k_adev *dev;
	int err;

	if (!card_avc8k) {
		printk(KERN_ERR "BUG: AVC8000 can't find device struct."
				" Can't proceed with open\n");
		return -ENODEV;
	}
	dev = card_avc8k->dev;

	daprintk(DPRT_LEVEL0, dev, "%s()\n", __func__);

	mutex_lock(&dev->lock);

	dev->read_offset = 0;

	mutex_unlock(&dev->lock);

	pcm = kzalloc(sizeof(*pcm), GFP_KERNEL);
	if (pcm == NULL)
		return -ENOMEM;

	pcm->dev=dev;

	spin_lock_init(&pcm->lock);

	pcm->substream = substream;
	runtime->private_data = pcm;
	runtime->private_free = snd_card_avc8k_runtime_free;
	runtime->hw = snd_card_avc8k_capture;

	err = snd_pcm_hw_constraint_integer(runtime,
						SNDRV_PCM_HW_PARAM_PERIODS);
	if (err < 0)
		return err;

	err = snd_pcm_hw_constraint_step(runtime, 0,
						SNDRV_PCM_HW_PARAM_PERIODS, 2);
	if (err < 0)
		return err;

	return 0;
}

/*
 * page callback (needed for mmap)
 */

static struct page *snd_card_avc8k_page(struct snd_pcm_substream *substream,
					unsigned long offset)
{
	void *pageptr = substream->runtime->dma_area + offset;

	snd_card_avc8k_t *card_avc8k = snd_pcm_substream_chip(substream);
	struct avc8k_adev *dev = card_avc8k->dev;

	daprintk(DPRT_LEVEL0, dev, "%s()\n", __func__);

	return vmalloc_to_page(pageptr);
}

/*
 * ALSA capture callbacks definition
 */

static struct snd_pcm_ops snd_card_avc8k_capture_ops = {
	.open =			snd_card_avc8k_capture_open,
	.close =		snd_card_avc8k_capture_close,
	.ioctl =		snd_pcm_lib_ioctl,
	.hw_params =	snd_card_avc8k_hw_params,
	.hw_free =		snd_card_avc8k_hw_free,
	.prepare =		snd_card_avc8k_capture_prepare,
	.trigger =		snd_card_avc8k_capture_trigger,
	.pointer =		snd_card_avc8k_capture_pointer,
	.page =			snd_card_avc8k_page,
};

/*
 * ALSA PCM setup
 *
 *   Called when initializing the board. Sets up the name and hooks up
 *  the callbacks
 *
 */

static int snd_card_avc8k_pcm(snd_card_avc8k_t *card_avc8k, long device)
{
	struct snd_pcm *pcm;
	int err;

	daprintk(DPRT_LEVEL0, card_avc8k->dev, "%s()\n", __func__);

	if ((err = snd_pcm_new(card_avc8k->card, "AVC8000 PCM", device, 0, 1, &pcm)) < 0)
		return err;
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, &snd_card_avc8k_capture_ops);
	pcm->private_data = card_avc8k;
	pcm->info_flags = 0;
	strcpy(pcm->name, "AVC8000 PCM");

	return 0;
}

static void snd_avc8k_free(struct snd_card * card)
{
    snd_card_avc8k_t *card_avc8k = (snd_card_avc8k_t*)card->private_data;
    struct avc8k_adev* dev = card_avc8k->dev;

    daprintk(DPRT_LEVEL0, dev, "%s()\n", __func__);
}

/*
 * ALSA initialization
 *
 *   Called by the init routine, once for each avc8k device present,
 *  it creates the basic structures and registers the ALSA devices
 *
 */

int avc8k_audio_create(struct avc8k_adev *dev)
{

	struct snd_card *card = NULL;
	snd_card_avc8k_t *card_avc8k;
	int err;
	char name[32];

	daprintk(DPRT_LEVEL0, dev, "%s()\n", __func__);

	if(avc8k_audio_nr > (SNDRV_CARDS-2))//7)//SNDRV_CARDS)
		return -ENODEV;
	sprintf(name, "AVC8000%li", avc8k_audio_nr);

#if(LINUX_VERSION_CODE >= KERNEL_VERSION(3,16,0))
	err = snd_card_new(NULL, -1, name, THIS_MODULE,
				sizeof(snd_card_avc8k_t), &card);
	if (err < 0)
		return err;
#elif(LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30))
	err = snd_card_create(-1, name, THIS_MODULE,
			      sizeof(snd_card_avc8k_t), &card);
	if (err < 0)
		return err;
#else
	card = snd_card_new(-2, name, THIS_MODULE, sizeof(snd_card_avc8k_t));
	if (card == NULL)
		return -ENOMEM;
#endif

	strcpy(card->driver, name);

	/* Card "creation" */

	card->private_free = snd_avc8k_free;
	card_avc8k = (snd_card_avc8k_t *) card->private_data;

	spin_lock_init(&card_avc8k->lock);

	card_avc8k->dev = dev;
	card_avc8k->card = card;
	dev->card = card_avc8k;

	mutex_init(&dev->lock);

	if ((err = snd_card_avc8k_pcm(card_avc8k, 0)) < 0)
		goto __nodev;

	snd_card_set_dev(card, &dev->chip->pci->dev);

	/* End of "creation" */

	strcpy(card->shortname, name);
	sprintf(card->longname, "%s at 0x%p irq %d",
		dev->chip->name, dev->chip->bmmio, dev->chip->pci->irq);

	daprintk(1, dev, "alsa: %s registered as card %d\n",card->longname,dev->channel_id);

	if ((err = snd_card_register(card)) == 0) {
	    avc8k_audio_nr++;
		return 0;
	}

__nodev:
	snd_card_free(card);
	return err;
}

int avc8k_audio_free(struct avc8k_adev *dev)
{
    if(dev->card) {
        snd_card_free(dev->card->card);
        dev->card = NULL;
    }

    avc8k_audio_nr--;

	return 1;
}
