/*
 * AVC8000 Copyright (C) 2012 Advanced Micro Peripherals Ltd
 * Driver for AVC8000 cards
 *
 * (c) 2009-10 liran <jlee@techwellinc.com.cn> [Techwell China]
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/init.h>
#include <linux/list.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kmod.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <asm/div64.h>

#include "avc8000-reg.h"
#include "avc8000.h"
#include "avc8000-device.h"

MODULE_DESCRIPTION("Driver for AVC8000 cards");
MODULE_AUTHOR("AMP");
MODULE_LICENSE("GPL");

unsigned int avc8k_debug;
module_param(avc8k_debug, int, 0644);
MODULE_PARM_DESC(avc8k_debug, "enable debug messages");

unsigned int avc8k_dmamode = AVC8K68XX_DMA_MODE_BLOCK;
module_param(avc8k_dmamode, int, 0644);
MODULE_PARM_DESC(avc8k_dmamode, "avc8000 dma mode");

unsigned int avc8k_audio = 0;//1;
module_param(avc8k_audio, int, 0644);
MODULE_PARM_DESC(avc8k_audio, "avc8000 audio enable");

unsigned int avc8k_restart_timer = 1;
module_param(avc8k_restart_timer, int, 0644);
MODULE_PARM_DESC(avc8k_restart_timer, "avc8000 dma restart timer enable");

unsigned int avc8k_bcsi = 0;
module_param(avc8k_bcsi, int, 0644);
MODULE_PARM_DESC(avc8k_bcsi, "avc8000 driver for BCSI enable");

static unsigned int avc8k_chipcount = 0;

static DEFINE_MUTEX(avc8k_chiplist_lock);
LIST_HEAD(avc8k_chiplist);


void avc8k_dmabuf_free(struct avc8k_chip *chip,
		       struct avc8k_dmabuf *buf)
{
    struct pci_dev *pci = chip->pci;

	if (NULL == buf->cpu)
		return;
	pci_free_consistent(pci, buf->size, buf->cpu, buf->dma);

    dprintk(DPRT_LEVEL0, chip, " dmamem free dma=%lx cpu=%p size=%d\n",
           (unsigned long)buf->dma, buf->cpu, buf->size);

	memset(buf,0,sizeof(*buf));
}

int avc8k_dmabuf_alloc(struct avc8k_chip *chip,
		       struct avc8k_dmabuf *buf,
		       unsigned int size)
{
    struct pci_dev *pci = chip->pci;
	__le32 *cpu;
	dma_addr_t dma = 0;

	if (NULL != buf->cpu && buf->size < size)
		avc8k_dmabuf_free(chip,buf);
	if (NULL == buf->cpu) {
	    size = PAGE_ALIGN(size);
		cpu = pci_alloc_consistent(pci, size, &dma);
		if (NULL == cpu) {
            dprintk(1, chip, "dmamem alloc failed size=%d\n", size);
			return -ENOMEM;
		}
		buf->cpu  = cpu;
		buf->dma  = dma;
		buf->size = size;
        dprintk(DPRT_LEVEL0, chip, "dmamem alloc dma=%lx cpu=%p size=%d\n",
           (unsigned long)dma, cpu, size);
	}
	memset(buf->cpu,0,buf->size);
	return 0;
}

#if(LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0))
void avc8k_pgtable_free(struct pci_dev *pci, struct avc8k_pgtable *pt)
{
	if (NULL == pt->cpu)
		return;
	pci_free_consistent(pci, pt->size, pt->cpu, pt->dma);
	pt->cpu = NULL;	
}
#endif

static void avc8k_shutdown(struct avc8k_chip *chip)
{
	/* Disable Video DMA */
	/* Disable Audio activity */
	/* Disable Interrupts */

    avc8k_dev_uninit( chip );
}

void avc8k_reset(struct avc8k_chip *chip)
{
	dprintk(1, chip, "%s()\n", __func__);

	avc8k_dev_uninit( chip );
    avc8k_dev_init( chip );
}

static int get_resources(struct avc8k_chip *chip)
{
	if (request_mem_region(pci_resource_start(chip->pci, 0),
			       pci_resource_len(chip->pci, 0),
			       chip->name))
		return 0;

	printk(KERN_ERR "%s: can't get MMIO memory @ 0x%llx\n",
		chip->name, (unsigned long long)pci_resource_start(chip->pci, 0));

	return -EBUSY;
}

static void avc8k_assist_timeout(unsigned long data)
{
	struct avc8k_chip* chip = (struct avc8k_chip*)data;
	int i=0;

	for(i=0; i<chip->vid_count; i++) {
		avc8k_video_status(chip->vid_dev[i]);
	}

	mod_timer(&chip->tm_assist, jiffies+BUFFER_TIMEOUT*2);
}

static int avc8k_setup(struct avc8k_chip *chip)
{
    struct avc8k_vdev *vdev;
    struct avc8k_adev *adev;
    int i = 0;

	chip->nr = avc8k_chipcount++;
	sprintf(chip->name, "avc8k[%d]", chip->nr);

	mutex_lock(&avc8k_chiplist_lock);
	list_add_tail(&chip->chiplist, &avc8k_chiplist);
	mutex_unlock(&avc8k_chiplist_lock);

	if (get_resources(chip) < 0) {
		printk(KERN_ERR "%s No more PCIe resources for "
		       "subsystem: %04x:%04x\n",
		       chip->name, chip->pci->subsystem_vendor,
		       chip->pci->subsystem_device);

		avc8k_chipcount--;
		return -ENODEV;
	}

	/* PCIe stuff */
	chip->bmmio = ioremap(pci_resource_start(chip->pci, 0),
			     pci_resource_len(chip->pci, 0));

	printk(KERN_INFO "%s: subsystem: %04x:%04x\n",
	       chip->name, chip->pci->subsystem_vendor,
	       chip->pci->subsystem_device);

	chip->vid_count = AVC8K68XX_DECODER_COUNT(chip->pci->device);

    

	/* init hardware */
	avc8k_reset(chip);
	if(chip->pci->device != PCI_DEVICE_ID_AVC8K6869)  // for avc8000x
	{
		avc8k_i2c_register(chip);
		chip->vid_count += avc8k_dev_init_286x(chip);
	}
	chip->aud_count  = chip->vid_count;
	GPIO_CTAL(chip,CLK_REG);

	for( i=0; i<chip->vid_count; i++ ) {
        vdev = kzalloc(sizeof(struct avc8k_vdev), GFP_KERNEL);
        if (NULL == vdev) {
            return -ENOMEM;
        }

        vdev->chip = chip;
        vdev->channel_id = i;
        chip->vid_dev[i]= vdev;

        if (avc8k_video_register(vdev) < 0) {
            printk(KERN_ERR "%s: %s() Failed to register "
                "video adapters %d\n", chip->name, __func__, i);
            return -ENOMEM;
        }
		
		if( avc8k_dmabuf_alloc(chip, &chip->dma_desc[AVC8K68XX_DMA_DESC_P], AVC8K68XX_DMA_DESC_LEN)<0 ) {
			printk(KERN_INFO "%s: allocate dma p-buffer descriptor memory failed\n", chip->name);
			return -ENOMEM;
		}
		if( avc8k_dmabuf_alloc(chip, &chip->dma_desc[AVC8K68XX_DMA_DESC_B], AVC8K68XX_DMA_DESC_LEN)<0 ) {
			printk(KERN_INFO "%s: allocate dma b-buffer descriptor memory failed\n", chip->name);
			return -ENOMEM;
		}
			avc8k_dev_set_pbdesc(vdev);
			
		if(avc8k_dmabuf_alloc(chip, &chip->dma_bouncebuffer, 720*576*3)<0)
		{
			printk(KERN_INFO "%s: allocate bounce buffer failed\n", chip->name);
			return -ENOMEM;
		}
    }
	

    if(avc8k_audio) {
        for( i=0; i<chip->aud_count; i++ ) {
            adev = kzalloc(sizeof(struct avc8k_adev), GFP_KERNEL);
            if (NULL == adev) {
                return -ENOMEM;
            }

            adev->chip = chip;
            adev->channel_id = i;
            chip->aud_dev[i]= adev;

            if (avc8k_audio_create(adev) < 0) {
                printk(KERN_ERR "%s: %s() Failed to create "
                    "audio adapters %d\n", chip->name, __func__, i);
                kfree( adev );
                chip->aud_dev[i] = NULL;
                break;
            }
        }
    }

	if(chip->vid_count > 0) {
		chip->tm_assist.function = avc8k_assist_timeout;
		chip->tm_assist.data = (unsigned long)chip;
		init_timer(&chip->tm_assist);
		mod_timer(&chip->tm_assist, jiffies+BUFFER_TIMEOUT*2);
	}

	return 0;
}

static void avc8k_unregister(struct avc8k_chip *chip)
{
    int i=0;

	release_mem_region(pci_resource_start(chip->pci, 0),
			   pci_resource_len(chip->pci, 0));

	if(chip->vid_count > 0) {
		del_timer(&chip->tm_assist);
	}

    for( i=0; i<chip->vid_count; i++ ) {
        if(chip->vid_dev[i]) {
            avc8k_video_unregister( chip->vid_dev[i] );
            avc8k_dmabuf_free(chip, &chip->dma_bouncebuffer);
			avc8k_dmabuf_free( chip, &chip->dma_desc[AVC8K68XX_DMA_DESC_P] );
			avc8k_dmabuf_free( chip, &chip->dma_desc[AVC8K68XX_DMA_DESC_B] );
            kfree( chip->vid_dev[i] );
            chip->vid_dev[i] = NULL;
        }
    }

    for( i=0; i<chip->aud_count; i++ ) {
        if(chip->aud_dev[i]) {
            avc8k_audio_free( chip->aud_dev[i] );
            kfree( chip->aud_dev[i] );
            chip->aud_dev[i] = NULL;
        }
    }

//    avc8k_dmabuf_free( chip, &chip->dma_desc[AVC8K68XX_DMA_DESC_P] );
//    avc8k_dmabuf_free( chip, &chip->dma_desc[AVC8K68XX_DMA_DESC_B] );

	if(chip->pci->device != PCI_DEVICE_ID_AVC8K6869)  // for avc8000x
	{
		avc8k_i2c_unregister(chip);
	}
	iounmap(chip->bmmio);
}

static irqreturn_t avc8k_irq(int irq, void *chip_id)
{
	struct avc8k_chip *chip = chip_id;
	struct avc8k_vdev *vdev = NULL;
	struct avc8k_adev *adev = NULL;
	int handled = 0;
	int i;
	u32 dma_status = avc8k_read( AVC8K68XX_R_DMA_INT_STATUS );
	u32 pb_status  = avc8k_read( AVC8K68XX_R_DMA_PB_STATUS );
	u32 dma_error  = avc8k_read( AVC8K68XX_R_DMA_INT_ERROR );
	u32 dma_cmd    = avc8k_read( AVC8K68XX_R_DMA_CMD );

    if( !(dma_cmd & AVC8K68XX_R_DMA_CMD_RESETALL) ) {
        return IRQ_RETVAL(handled);
    }

    if( (dma_status == 0) && (dma_error == 0) ) {
        return IRQ_RETVAL(handled);
    }
    
	for( i=0; i<chip->vid_count; i++ ) {
		if( (dma_status & AVC8K68XX_R_DMA_INT_STATUS_DMA(i))/* || (dma_error & AVC8K68XX_R_DMA_VINT_FIFO(i))*/ ) {
			vdev = chip->vid_dev[i];
			avc8k_video_irq( vdev, dma_status, pb_status );
        }
        else if( dma_error & AVC8K68XX_R_DMA_VINT_FIFO(i) ) {
        	vdev = chip->vid_dev[i];
			spin_lock(&vdev->slock);
			avc8k_dev_run( vdev, false );
			spin_unlock(&vdev->slock);
	    }
    }
    for( i=0; i<chip->aud_count; i++ ) {
    	if( dma_status & AVC8K68XX_R_DMA_INT_STATUS_DMA(i+AVC8K68XX_AUDIO_BEGIN) ) {
        	adev = chip->aud_dev[i];
            avc8k_audio_irq( adev, dma_status, pb_status );
        }
    }
//  if( dma_error & AVC8K68XX_R_DMA_VINT_ERROR ) {
//  	avc8k_video_daemon( chip, dma_error );
//  }

    if( dma_status & AVC8K68XX_R_DMA_INT_STATUS_TOUT ) {
        //DMA time out
        avc8k_video_resetdma( chip, dma_cmd );
    }

    handled = 1;

	return IRQ_RETVAL(handled);
}

#if(LINUX_VERSION_CODE < KERNEL_VERSION(2,6,38))
static int __devinit avc8k_initchip(struct pci_dev *pci_dev,
				     const struct pci_device_id *pci_id)
#else
static int avc8k_initchip(struct pci_dev *pci_dev,
				     const struct pci_device_id *pci_id)
#endif
{
	struct avc8k_chip *chip;
	int err;
	u8  pci_rev,pci_lat;

	chip = kzalloc(sizeof(*chip), GFP_KERNEL);
	if (NULL == chip)
		return -ENOMEM;

	mutex_init(&chip->lock);

#if(LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30))
	err = v4l2_device_register(&pci_dev->dev, &chip->v4l2_dev);
	if (err < 0)
		goto fail_free;
#endif

	/* pci init */
	chip->pci = pci_dev;
	if (pci_enable_device(pci_dev)) {
		err = -EIO;
		goto fail_unreg;
	}

	if (avc8k_setup(chip) < 0) {
		err = -EINVAL;
		goto fail_unreg;
	}

	/* print pci info */
	pci_read_config_byte(pci_dev, PCI_CLASS_REVISION, &pci_rev);
	pci_read_config_byte(pci_dev, PCI_LATENCY_TIMER,  &pci_lat);
	printk(KERN_INFO "%s: found at %s, bus: %d, rev: %d, irq: %d, "
		"latency: %d, mmio: 0x%llx, device:0x%x\n", chip->name,
	       pci_name(pci_dev), pci_dev->bus->number, pci_rev, pci_dev->irq, pci_lat,
           (unsigned long long)pci_resource_start(pci_dev, 0),pci_dev->device);

	pci_set_master(pci_dev);
#if(LINUX_VERSION_CODE < KERNEL_VERSION(4,4,0))
	if (!pci_dma_supported(pci_dev, 0xffffffff)) {
#else
	err = pci_set_dma_mask(pci_dev, 0xffffffff);
	if (err) {
#endif
		printk("%s: Oops: no 32bit PCI DMA ???\n", chip->name);
#if(LINUX_VERSION_CODE < KERNEL_VERSION(4,4,0))
		err = -EIO;
#endif
		goto fail_irq;
	}

#if(LINUX_VERSION_CODE < KERNEL_VERSION(4,1,0))
	err = request_irq(pci_dev->irq, avc8k_irq,
			  IRQF_SHARED | IRQF_DISABLED, chip->name, chip);
#else
	err = request_irq(pci_dev->irq, avc8k_irq, IRQF_SHARED, chip->name, chip);
#endif
	if (err < 0) {
		printk(KERN_ERR "%s: can't get IRQ %d\n",
		       chip->name, pci_dev->irq);
		goto fail_irq;
	}



#if(LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,30))
	pci_set_drvdata(pci_dev, chip);
#endif

	return 0;

fail_irq:
	avc8k_unregister(chip);
fail_unreg:
#if(LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30))
	v4l2_device_unregister(&chip->v4l2_dev);
fail_free:
#endif
	kfree(chip);
	return err;
}


#if(LINUX_VERSION_CODE < KERNEL_VERSION(2,6,38))
static void __devexit avc8k_finichip(struct pci_dev *pci_dev)
#else
static void avc8k_finichip(struct pci_dev *pci_dev)
#endif
{
#if(LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30))
	struct v4l2_device *v4l2_dev = pci_get_drvdata(pci_dev);
	struct avc8k_chip *chip = to_avc8k(v4l2_dev);
#else
	struct avc8k_chip *chip = pci_get_drvdata(pci_dev);
#endif

	avc8k_shutdown(chip);

	pci_disable_device(pci_dev);
#if(LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,30))
	pci_set_drvdata(pci_dev, NULL);
#endif

	/* unregister stuff */
	free_irq(pci_dev->irq, chip);

	mutex_lock(&avc8k_chiplist_lock);
	list_del(&chip->chiplist);
	mutex_unlock(&avc8k_chiplist_lock);

	avc8k_unregister(chip);
#if(LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30))
	v4l2_device_unregister(v4l2_dev);
#endif
	kfree(chip);
}

static struct pci_device_id avc8k_pci_tbl[] = {
	{
		
		.vendor       = 0x1797,
		.device       = PCI_DEVICE_ID_AVC8K6864,
		.subvendor    = PCI_ANY_ID,
		.subdevice    = PCI_ANY_ID,
	},
	{
		
		.vendor       = 0x1797,
		.device       = PCI_DEVICE_ID_AVC8K6869,
		.subvendor    = PCI_ANY_ID,
		.subdevice    = PCI_ANY_ID,
	},
	{
		/* --- end of list --- */
	}
};
MODULE_DEVICE_TABLE(pci, avc8k_pci_tbl);

static struct pci_driver avc8k_pci_driver = {
	.name     = "avc8k",
	.id_table = avc8k_pci_tbl,
	.probe    = avc8k_initchip,
#if(LINUX_VERSION_CODE < KERNEL_VERSION(2,6,38))
	.remove   = __devexit_p(avc8k_finichip),
#else
	.remove   = avc8k_finichip,
#endif
	/* TODO */
	.suspend  = NULL,
	.resume   = NULL,
};

static int avc8k_init(void)
{
	printk(KERN_INFO "avc8000 driver version %d.%d.%d loaded\n",
	       (AVC8K_VERSION_CODE >> 16) & 0xff,
	       (AVC8K_VERSION_CODE >>  8) & 0xff,
	       AVC8K_VERSION_CODE & 0xff);
    printk(KERN_INFO "avc8000 driver dma_mode=%x, debug=%x\n", avc8k_dmamode, avc8k_debug);

	return pci_register_driver(&avc8k_pci_driver);
}

static void avc8k_fini(void)
{
	pci_unregister_driver(&avc8k_pci_driver);

	printk(KERN_INFO "avc8000 driver version %d.%d.%d unloaded\n",
	       (AVC8K_VERSION_CODE >> 16) & 0xff,
	       (AVC8K_VERSION_CODE >>  8) & 0xff,
	       AVC8K_VERSION_CODE & 0xff);
}

module_init(avc8k_init);
module_exit(avc8k_fini);
